@@ -1,4 +1,3 @@
-import deprecation
 import logging
 import os
 from typing import Iterator, List, Optional, Union
@@ -15,8 +14,8 @@ from .common import (
 )
 from .entity import (
     Bar, Entity, Account, AccountConfigurations, AccountActivity,
-    Asset, Order, Position, BarSet, Clock, Calendar,
-    Aggs, Trade, Quote, Watchlist, PortfolioHistory
+    Asset, Order, Position, Clock, Calendar,
+    Trade, Quote, Watchlist, PortfolioHistory
 )
 from .entity_v2 import (
     BarV2, BarsV2, LatestBarsV2, LatestQuotesV2, LatestTradesV2,
@@ -80,6 +79,8 @@ class TimeFrameUnit(Enum):
     Minute = "Min"
     Hour = "Hour"
     Day = "Day"
+    Week = "Week"
+    Month = "Month"
 
 
 class TimeFrame:
@@ -126,11 +127,21 @@ class TimeFrame:
         if unit == TimeFrameUnit.Hour and amount > 23:
             raise ValueError("Hour units can only be used with amounts 1-23")
 
+        if unit in (TimeFrameUnit.Day, TimeFrameUnit.Week) and amount != 1:
+            raise ValueError(
+                "Day and Week units can only be used with amount 1")
+
+        if unit == TimeFrameUnit.Month and amount not in (1, 2, 3, 6, 12):
+            raise ValueError(
+                "Month units can only be used with amount 1, 2, 3, 6 and 12")
+
 
 # These are kept for backwards compatibility
 TimeFrame.Minute = TimeFrame(1, TimeFrameUnit.Minute)
 TimeFrame.Hour = TimeFrame(1, TimeFrameUnit.Hour)
 TimeFrame.Day = TimeFrame(1, TimeFrameUnit.Day)
+TimeFrame.Week = TimeFrame(1, TimeFrameUnit.Week)
+TimeFrame.Month = TimeFrame(1, TimeFrameUnit.Month)
 
 
 class Sort(Enum):
@@ -250,8 +261,12 @@ class REST(object):
     def delete(self, path, data=None):
         return self._request('DELETE', path, data)
 
-    def data_get(self, path, data=None, api_version='v1'):
+    def data_get(self, path, data=None,
+                 feed: Optional[str] = None, api_version='v1'):
         base_url: URL = get_data_url()
+        if feed:
+            data = data or {}
+            data['feed'] = feed
         return self._request(
             'GET', path, data, base_url=base_url, api_version=api_version,
         )
@@ -298,7 +313,8 @@ class REST(object):
                     direction: str = None,
                     params=None,
                     nested: bool = None,
-                    symbols: List[str] = None
+                    symbols: List[str] = None,
+                    side: str = None
                     ) -> Orders:
         """
         Get a list of orders
@@ -312,6 +328,7 @@ class REST(object):
         :param params: refer to documentation
         :param nested: should the data be nested like json
         :param symbols: list of str (symbols)
+        :param side: Lets you filter to only 'buy' or 'sell' orders
         """
         if params is None:
             params = dict()
@@ -327,6 +344,8 @@ class REST(object):
             params['status'] = status
         if nested is not None:
             params['nested'] = nested
+        if side is not None:
+            params['side'] = side
         if symbols is not None:
             params['symbols'] = ",".join(symbols)
         url = '/orders'
@@ -530,92 +549,6 @@ class REST(object):
         resp = self.get('/assets/{}'.format(symbol))
         return self.response_wrapper(resp, Asset)
 
-    @deprecation.deprecated(deprecated_in="v1.0.0",
-                            details="Use get_bars instead")
-    def get_barset(self,
-                   symbols,
-                   timeframe: str,
-                   limit: int = None,
-                   start: str = None,
-                   end: str = None,
-                   after: str = None,
-                   until: str = None) -> BarSet:
-        """
-        read the documentation here:
-        https://alpaca.markets/docs/api-documentation/api-v2/market-data/bars/
-        Get BarSet(dict[str]->list[Bar])
-        :param symbols: The parameter symbols can be either a comma-split
-               string or a list of string. Each symbol becomes the key of the
-               returned value.
-        :param timeframe: One of minute, 1Min, 5Min, 15Min, day or 1D. minute
-               is an alias of 1Min. Similarly, day is of 1D.
-        :param limit: The maximum number of bars per symbol. It can be between
-               1 and 1000. Default is 100.
-        :param start: ISO Format str, ex: '2019-04-15T09:30:00-04:00' or
-               '2019-04-15'
-        :param end: ISO Format str
-        :param after: ISO Format str
-        :param until: ISO Format str
-        :return: BarSet
-
-        note: start can't be used with after. end cannot be used with until.
-        """
-        if not isinstance(symbols, str):
-            symbols = ','.join(symbols)
-        params = {
-            'symbols': symbols,
-        }
-        if limit is not None:
-            params['limit'] = limit
-        if start is not None:
-            params['start'] = start
-        if end is not None:
-            params['end'] = end
-        if after is not None:
-            params['after'] = after
-        if until is not None:
-            params['until'] = until
-        resp = self.data_get('/bars/{}'.format(timeframe), params)
-        return self.response_wrapper(resp, BarSet)
-
-    @deprecation.deprecated(deprecated_in="v1.0.0",
-                            details="Use get_bars instead")
-    def get_aggs(self,
-                 symbol: str,
-                 multiplier: int,
-                 timespan: str,
-                 _from: str,
-                 to: str) -> Aggs:
-        """
-
-        :param symbol: str eg AAPL
-        :param multiplier: must be 1
-        :param timespan: day or minute
-        :param _from: yyyy-mm-dd
-        :param to: yyyy-mm-dd
-        :return:
-        """
-        resp = self.data_get('/aggs/ticker/{}/range/{}/{}/{}/{}'.format(
-            symbol, multiplier, timespan, _from, to
-        ))
-        return self.response_wrapper(resp, Aggs)
-
-    @deprecation.deprecated(deprecated_in="v1.0.0",
-                            details="Use get_latest_trade instead")
-    def get_last_trade(self, symbol: str) -> Trade:
-        """
-        Get the last trade for the given symbol
-        """
-        resp = self.data_get('/last/stocks/{}'.format(symbol))
-        return self.response_wrapper(resp['last'], Trade)
-
-    @deprecation.deprecated(deprecated_in="v1.0.0",
-                            details="Use get_latest_quote instead")
-    def get_last_quote(self, symbol: str) -> Quote:
-        """Get the last quote for the given symbol"""
-        resp = self.data_get('/last_quote/stocks/{}'.format(symbol))
-        return self.response_wrapper(resp['last'], Quote)
-
     def _data_get(self,
                   endpoint: str,
                   symbol_or_symbols: Union[str, List[str]],
@@ -623,6 +556,7 @@ class REST(object):
                   endpoint_base: str = 'stocks',
                   resp_grouped_by_symbol: Optional[bool] = None,
                   page_limit: int = DATA_V2_MAX_LIMIT,
+                  feed: Optional[str] = None,
                   **kwargs):
         page_token = None
         total_items = 0
@@ -645,7 +579,8 @@ class REST(object):
                 data['symbols'] = ','.join(symbol_or_symbols)
             if endpoint:
                 path += f'/{endpoint}'
-            resp = self.data_get(path, data=data, api_version=api_version)
+            resp = self.data_get(path, data=data, feed=feed,
+                                 api_version=api_version)
             if not resp_grouped_by_symbol:
                 k = endpoint or endpoint_base
                 for item in resp.get(k, []) or []:
@@ -667,9 +602,10 @@ class REST(object):
                         start: Optional[str] = None,
                         end: Optional[str] = None,
                         limit: int = None,
+                        feed: Optional[str] = None,
                         raw=False) -> TradeIterator:
         trades = self._data_get('trades', symbol,
-                                start=start, end=end, limit=limit)
+                                start=start, end=end, limit=limit, feed=feed)
         for trade in trades:
             if raw:
                 yield trade
@@ -681,9 +617,10 @@ class REST(object):
                    start: Optional[str] = None,
                    end: Optional[str] = None,
                    limit: int = None,
+                   feed: Optional[str] = None,
                    ) -> TradesV2:
         trades = list(self.get_trades_iter(symbol,
-                                           start, end, limit, raw=True))
+                                           start, end, limit, feed, raw=True))
         return TradesV2(trades)
 
     def get_quotes_iter(self,
@@ -691,9 +628,10 @@ class REST(object):
                         start: Optional[str] = None,
                         end: Optional[str] = None,
                         limit: int = None,
+                        feed: Optional[str] = None,
                         raw=False) -> QuoteIterator:
         quotes = self._data_get('quotes', symbol,
-                                start=start, end=end, limit=limit)
+                                start=start, end=end, limit=limit, feed=feed)
         for quote in quotes:
             if raw:
                 yield quote
@@ -705,11 +643,13 @@ class REST(object):
                    start: Optional[str] = None,
                    end: Optional[str] = None,
                    limit: int = None,
+                   feed: Optional[str] = None,
                    ) -> QuotesV2:
-        quotes = list(self.get_quotes_iter(symbol,
-                                           start,
-                                           end,
-                                           limit,
+        quotes = list(self.get_quotes_iter(symbol=symbol,
+                                           start=start,
+                                           end=end,
+                                           limit=limit,
+                                           feed=feed,
                                            raw=True))
         return QuotesV2(quotes)
 
@@ -720,11 +660,12 @@ class REST(object):
                       end: Optional[str] = None,
                       adjustment: str = 'raw',
                       limit: int = None,
+                      feed: Optional[str] = None,
                       raw=False) -> BarIterator:
         bars = self._data_get('bars', symbol,
                               timeframe=timeframe,
                               adjustment=adjustment,
-                              start=start, end=end, limit=limit)
+                              start=start, end=end, limit=limit, feed=feed)
         for bar in bars:
             if raw:
                 yield bar
@@ -738,6 +679,7 @@ class REST(object):
                  end: Optional[str] = None,
                  adjustment: str = 'raw',
                  limit: int = None,
+                 feed: Optional[str] = None,
                  ) -> BarsV2:
         bars = list(self.get_bars_iter(symbol,
                                        timeframe,
@@ -745,53 +687,69 @@ class REST(object):
                                        end,
                                        adjustment,
                                        limit,
+                                       feed=feed,
                                        raw=True))
         return BarsV2(bars)
 
-    def get_latest_bar(self, symbol: str) -> BarV2:
+    def get_latest_bar(self, symbol: str, feed: Optional[str] = None) -> BarV2:
         resp = self.data_get(
             '/stocks/{}/bars/latest'.format(symbol),
+            feed=feed,
             api_version='v2')
         return self.response_wrapper(resp['bar'], BarV2)
 
-    def get_latest_bars(self, symbols: List[str]) -> LatestBarsV2:
+    def get_latest_bars(self, symbols: List[str],
+                        feed: Optional[str] = None) -> LatestBarsV2:
         resp = self.data_get(
-            f'/stocks/bars/latest?symbols={",".join(symbols)}',
+            f'/stocks/bars/latest?symbols={_join_with_commas(symbols)}',
+            feed=feed,
             api_version='v2')
         return self.response_wrapper(resp['bars'], LatestBarsV2)
 
-    def get_latest_trade(self, symbol: str) -> TradeV2:
+    def get_latest_trade(self, symbol: str,
+                         feed: Optional[str] = None) -> TradeV2:
         resp = self.data_get(
             '/stocks/{}/trades/latest'.format(symbol),
+            feed=feed,
             api_version='v2')
         return self.response_wrapper(resp['trade'], TradeV2)
 
-    def get_latest_trades(self, symbols: List[str]) -> LatestTradesV2:
+    def get_latest_trades(self, symbols: List[str],
+                          feed: Optional[str] = None) -> LatestTradesV2:
         resp = self.data_get(
-            f'/stocks/trades/latest?symbols={",".join(symbols)}',
+            f'/stocks/trades/latest?symbols={_join_with_commas(symbols)}',
+            feed=feed,
             api_version='v2')
         return self.response_wrapper(resp['trades'], LatestTradesV2)
 
-    def get_latest_quote(self, symbol: str) -> QuoteV2:
+    def get_latest_quote(self, symbol: str,
+                         feed: Optional[str] = None) -> QuoteV2:
         resp = self.data_get(
             '/stocks/{}/quotes/latest'.format(symbol),
+            feed=feed,
             api_version='v2')
         return self.response_wrapper(resp['quote'], QuoteV2)
 
-    def get_latest_quotes(self, symbols: List[str]) -> LatestQuotesV2:
+    def get_latest_quotes(self, symbols: List[str],
+                          feed: Optional[str] = None) -> LatestQuotesV2:
         resp = self.data_get(
-            f'/stocks/quotes/latest?symbols={",".join(symbols)}',
+            f'/stocks/quotes/latest?symbols={_join_with_commas(symbols)}',
+            feed=feed,
             api_version='v2')
         return self.response_wrapper(resp['quotes'], LatestQuotesV2)
 
-    def get_snapshot(self, symbol: str) -> SnapshotV2:
+    def get_snapshot(self, symbol: str,
+                     feed: Optional[str] = None) -> SnapshotV2:
         resp = self.data_get('/stocks/{}/snapshot'.format(symbol),
+                             feed=feed,
                              api_version='v2')
         return self.response_wrapper(resp, SnapshotV2)
 
-    def get_snapshots(self, symbols: List[str]) -> SnapshotsV2:
+    def get_snapshots(self, symbols: List[str],
+                      feed: Optional[str] = None) -> SnapshotsV2:
         resp = self.data_get(
-            '/stocks/snapshots?symbols={}'.format(','.join(symbols)),
+            '/stocks/snapshots?symbols={}'.format(_join_with_commas(symbols)),
+            feed=feed,
             api_version='v2')
         return self.response_wrapper(resp, SnapshotsV2)
 
@@ -879,6 +837,22 @@ class REST(object):
         return BarsV2(list(self.get_crypto_bars_iter(
             symbol, timeframe, start, end, limit, exchanges, raw=True)))
 
+    def get_latest_crypto_bar(self, symbol: str, exchange: str) -> BarV2:
+        resp = self.data_get(
+            '/crypto/{}/bars/latest'.format(symbol),
+            data={'exchange': exchange},
+            api_version='v1beta1')
+        return self.response_wrapper(resp['bar'], BarV2)
+
+    def get_latest_crypto_bars(self,
+                               symbols: List[str],
+                               exchange: str) -> LatestBarsV2:
+        resp = self.data_get(
+            '/crypto/bars/latest',
+            data={'exchange': exchange, 'symbols': _join_with_commas(symbols)},
+            api_version='v1beta1')
+        return self.response_wrapper(resp['bars'], LatestBarsV2)
+
     def get_latest_crypto_trade(self, symbol: str, exchange: str) -> TradeV2:
         resp = self.data_get(
             '/crypto/{}/trades/latest'.format(symbol),
@@ -886,6 +860,15 @@ class REST(object):
             api_version='v1beta1')
         return self.response_wrapper(resp['trade'], TradeV2)
 
+    def get_latest_crypto_trades(self,
+                                 symbols: List[str],
+                                 exchange: str) -> LatestTradesV2:
+        resp = self.data_get(
+            '/crypto/trades/latest',
+            data={'exchange': exchange, 'symbols': _join_with_commas(symbols)},
+            api_version='v1beta1')
+        return self.response_wrapper(resp['trades'], LatestTradesV2)
+
     def get_latest_crypto_quote(self, symbol: str, exchange: str) -> QuoteV2:
         resp = self.data_get(
             '/crypto/{}/quotes/latest'.format(symbol),
@@ -893,25 +876,56 @@ class REST(object):
             api_version='v1beta1')
         return self.response_wrapper(resp['quote'], QuoteV2)
 
+    def get_latest_crypto_quotes(self,
+                                 symbols: List[str],
+                                 exchange: str) -> LatestQuotesV2:
+        resp = self.data_get(
+            '/crypto/quotes/latest',
+            data={'exchange': exchange, 'symbols': _join_with_commas(symbols)},
+            api_version='v1beta1')
+        return self.response_wrapper(resp['quotes'], LatestQuotesV2)
+
     def get_latest_crypto_xbbo(self,
                                symbol: str,
                                exchanges: Optional[List[str]] = None,
                                ) -> QuoteV2:
         params = {}
         if exchanges:
-            params['exchanges'] = ','.join(exchanges)
+            params['exchanges'] = _join_with_commas(exchanges)
         resp = self.data_get(
             '/crypto/{}/xbbo/latest'.format(symbol),
             data=params,
             api_version='v1beta1')
         return self.response_wrapper(resp['xbbo'], QuoteV2)
 
+    def get_latest_crypto_xbbos(self,
+                                symbols: List[str],
+                                exchanges: Optional[List[str]] = None,
+                                ) -> QuoteV2:
+        params = {'symbols': _join_with_commas(symbols)}
+        if exchanges:
+            params['exchanges'] = _join_with_commas(exchanges)
+        resp = self.data_get(
+            '/crypto/xbbos/latest',
+            data=params,
+            api_version='v1beta1')
+        return self.response_wrapper(resp['xbbos'], QuotesV2)
+
     def get_crypto_snapshot(self, symbol: str, exchange: str) -> SnapshotV2:
         resp = self.data_get('/crypto/{}/snapshot'.format(symbol),
                              data={'exchange': exchange},
                              api_version='v1beta1')
         return self.response_wrapper(resp, SnapshotV2)
 
+    def get_latest_crypto_snapshots(self,
+                                    symbols: List[str],
+                                    exchange: str) -> SnapshotsV2:
+        resp = self.data_get(
+            '/crypto/snapshots',
+            data={'exchange': exchange, 'symbols': _join_with_commas(symbols)},
+            api_version='v1beta1')
+        return self.response_wrapper(resp['snapshots'], SnapshotsV2)
+
     def get_news_iter(self,
                       symbol: Optional[Union[str, List[str]]] = None,
                       start: Optional[str] = None,
@@ -1137,3 +1151,9 @@ class REST(object):
             return obj
         else:
             return entity(obj)
+
+
+def _join_with_commas(lst: List[str]) -> str:
+    if isinstance(lst, str):
+        raise ValueError('expected list, str found')
+    return ','.join(lst)
