@@ -1,12 +1,8 @@
-"""
-Stream V2
-For historic reasons stream2.py contains the old api version.
-Don't get confused
-"""
 import asyncio
+from collections import defaultdict
 import logging
 import json
-from typing import List, Optional
+from typing import Dict, List, Optional
 import msgpack
 import re
 import websockets
@@ -22,6 +18,7 @@ from .entity_v2 import (
     luld_mapping_v2,
     cancel_error_mapping_v2,
     correction_mapping_v2,
+    orderbook_mapping_v2,
     Trade,
     Quote,
     Bar,
@@ -30,42 +27,60 @@ from .entity_v2 import (
     CancelErrorV2,
     CorrectionV2,
     NewsV2,
+    OrderbookV2,
 )
 
 log = logging.getLogger(__name__)
 
+# Default Params we pass to the websocket constructors
+WEBSOCKET_DEFAULTS = {
+    "ping_interval": 10,
+    "ping_timeout": 180,
+    "max_queue": 1024,
+}
+
 
 def _ensure_coroutine(handler):
     if not asyncio.iscoroutinefunction(handler):
         raise ValueError('handler must be a coroutine function')
 
 
-class _DataStream():
+class _DataStream:
     def __init__(self,
                  endpoint: str,
                  key_id: str,
                  secret_key: str,
-                 raw_data: bool = False) -> None:
+                 raw_data: bool = False,
+                 websocket_params: Optional[Dict] = None) -> None:
         self._endpoint = endpoint
         self._key_id = key_id
         self._secret_key = secret_key
         self._ws = None
         self._running = False
+        self._loop = None
         self._raw_data = raw_data
         self._stop_stream_queue = queue.Queue()
         self._handlers = {
-            'trades':    {},
-            'quotes':    {},
-            'bars':      {},
-            'dailyBars': {},
+            'trades':      {},
+            'quotes':      {},
+            'bars':        {},
+            'updatedBars': {},
+            'dailyBars':   {},
         }
         self._name = 'data'
         self._should_run = True
+        self._max_frame_size = 32768
+        self._websocket_params = websocket_params
+
+        if self._websocket_params is None:
+            self._websocket_params = WEBSOCKET_DEFAULTS
 
     async def _connect(self):
         self._ws = await websockets.connect(
             self._endpoint,
-            extra_headers={'Content-Type': 'application/msgpack'})
+            extra_headers={'Content-Type': 'application/msgpack'},
+            **self._websocket_params
+        )
         r = await self._ws.recv()
         msg = msgpack.unpackb(r)
         if msg[0]['T'] != 'success' or msg[0]['msg'] != 'connected':
@@ -136,7 +151,7 @@ class _DataStream():
                     quote_mapping_v2[k]: v
                     for k, v in msg.items() if k in quote_mapping_v2
                 })
-            elif msg_type in ('b', 'd'):
+            elif msg_type in ('b', 'u', 'd'):
                 result = Bar({
                     bar_mapping_v2[k]: v
                     for k, v in msg.items() if k in bar_mapping_v2
@@ -163,6 +178,11 @@ class _DataStream():
                 symbol, self._handlers['bars'].get('*', None))
             if handler:
                 await handler(self._cast(msg_type, msg))
+        elif msg_type == 'u':
+            handler = self._handlers['updatedBars'].get(
+                symbol, self._handlers['updatedBars'].get('*', None))
+            if handler:
+                await handler(self._cast(msg_type, msg))
         elif msg_type == 'd':
             handler = self._handlers['dailyBars'].get(
                 symbol, self._handlers['dailyBars'].get('*', None))
@@ -179,37 +199,32 @@ class _DataStream():
         for symbol in symbols:
             handlers[symbol] = handler
         if self._running:
-            asyncio.get_event_loop().run_until_complete(self._subscribe_all())
+            asyncio.run_coroutine_threadsafe(
+                self._subscribe_all(), self._loop
+            ).result()
 
     async def _subscribe_all(self):
-        if any(
-            v for k, v in self._handlers.items()
-            if k not in ("cancelErrors", "corrections")
-        ):
-            msg = {
-                k: tuple(v.keys())
-                for k, v in self._handlers.items()
-                if v
-            }
-            msg['action'] = 'subscribe'
-            await self._ws.send(msgpack.packb(msg))
+        msg = defaultdict(list)
+        for k, v in self._handlers.items():
+            if k not in ("cancelErrors", "corrections") and v:
+                for s in v.keys():
+                    msg[k].append(s)
+        msg['action'] = 'subscribe'
+        bs = msgpack.packb(msg)
+        frames = (bs[i:i+self._max_frame_size]
+                  for i in range(0, len(bs), self._max_frame_size))
+        await self._ws.send(frames)
 
     async def _unsubscribe(self,
                            trades=(),
                            quotes=(),
                            bars=(),
+                           updated_bars=(),
                            daily_bars=()):
-        if trades or quotes or bars or daily_bars:
-            await self._ws.send(
-                msgpack.packb({
-                    'action':    'unsubscribe',
-                    'trades':    trades,
-                    'quotes':    quotes,
-                    'bars':      bars,
-                    'dailyBars': daily_bars,
-                }))
+        raise NotImplementedError()
 
     async def _run_forever(self):
+        self._loop = asyncio.get_running_loop()
         # do not start the websocket connection until we subscribe to something
         while not any(
             v for k, v in self._handlers.items()
@@ -257,37 +272,58 @@ class _DataStream():
     def subscribe_bars(self, handler, *symbols):
         self._subscribe(handler, symbols, self._handlers['bars'])
 
+    def subscribe_updated_bars(self, handler, *symbols):
+        self._subscribe(handler, symbols, self._handlers['updatedBars'])
+
     def subscribe_daily_bars(self, handler, *symbols):
         self._subscribe(handler, symbols, self._handlers['dailyBars'])
 
     def unsubscribe_trades(self, *symbols):
         if self._running:
-            asyncio.get_event_loop().run_until_complete(
-                self._unsubscribe(trades=symbols))
+            asyncio.run_coroutine_threadsafe(
+                self._unsubscribe(trades=symbols),
+                self._loop).result()
         for symbol in symbols:
             del self._handlers['trades'][symbol]
 
     def unsubscribe_quotes(self, *symbols):
         if self._running:
-            asyncio.get_event_loop().run_until_complete(
-                self._unsubscribe(quotes=symbols))
+            asyncio.run_coroutine_threadsafe(
+                self._unsubscribe(quotes=symbols),
+                self._loop).result()
         for symbol in symbols:
             del self._handlers['quotes'][symbol]
 
     def unsubscribe_bars(self, *symbols):
         if self._running:
-            asyncio.get_event_loop().run_until_complete(
-                self._unsubscribe(bars=symbols))
+            asyncio.run_coroutine_threadsafe(
+                self._unsubscribe(bars=symbols),
+                self._loop).result()
         for symbol in symbols:
             del self._handlers['bars'][symbol]
 
+    def unsubscribe_updated_bars(self, *symbols):
+        if self._running:
+            asyncio.run_coroutine_threadsafe(
+                self._unsubscribe(updated_bars=symbols),
+                self._loop).result()
+        for symbol in symbols:
+            del self._handlers['updatedBars'][symbol]
+
     def unsubscribe_daily_bars(self, *symbols):
         if self._running:
-            asyncio.get_event_loop().run_until_complete(
-                self._unsubscribe(daily_bars=symbols))
+            asyncio.run_coroutine_threadsafe(
+                self._unsubscribe(daily_bars=symbols),
+                self._loop).result()
         for symbol in symbols:
             del self._handlers['dailyBars'][symbol]
 
+    def stop(self):
+        if self._loop.is_running():
+            asyncio.run_coroutine_threadsafe(
+                self.stop_ws(),
+                self._loop).result()
+
 
 class DataStream(_DataStream):
     def __init__(self,
@@ -295,12 +331,14 @@ class DataStream(_DataStream):
                  secret_key: str,
                  base_url: URL,
                  raw_data: bool,
-                 feed: str = 'iex'):
+                 feed: str = 'iex',
+                 websocket_params: Optional[Dict] = None):
         base_url = re.sub(r'^http', 'ws', base_url)
         super().__init__(endpoint=base_url + '/v2/' + feed,
                          key_id=key_id,
                          secret_key=secret_key,
                          raw_data=raw_data,
+                         websocket_params=websocket_params
                          )
         self._handlers['statuses'] = {}
         self._handlers['lulds'] = {}
@@ -363,19 +401,22 @@ class DataStream(_DataStream):
                            trades=(),
                            quotes=(),
                            bars=(),
+                           updated_bars=(),
                            daily_bars=(),
                            statuses=(),
                            lulds=()):
-        if trades or quotes or bars or daily_bars or statuses or lulds:
+        if (trades or quotes or bars or updated_bars or daily_bars or
+                statuses or lulds):
             await self._ws.send(
                 msgpack.packb({
-                    'action':    'unsubscribe',
-                    'trades':    trades,
-                    'quotes':    quotes,
-                    'bars':      bars,
-                    'dailyBars': daily_bars,
-                    'statuses':  statuses,
-                    'lulds':     lulds,
+                    'action':      'unsubscribe',
+                    'trades':      trades,
+                    'quotes':      quotes,
+                    'bars':        bars,
+                    'updatedBars': updated_bars,
+                    'dailyBars':   daily_bars,
+                    'statuses':    statuses,
+                    'lulds':       lulds,
                 }))
 
     def subscribe_statuses(self, handler, *symbols):
@@ -386,15 +427,17 @@ class DataStream(_DataStream):
 
     def unsubscribe_statuses(self, *symbols):
         if self._running:
-            asyncio.get_event_loop().run_until_complete(
-                self._unsubscribe(statuses=symbols))
+            asyncio.run_coroutine_threadsafe(
+                self._unsubscribe(statuses=symbols),
+                self._loop).result()
         for symbol in symbols:
             del self._handlers['statuses'][symbol]
 
     def unsubscribe_lulds(self, *symbols):
         if self._running:
-            asyncio.get_event_loop().run_until_complete(
-                self._unsubscribe(lulds=symbols))
+            asyncio.run_coroutine_threadsafe(
+                self._unsubscribe(lulds=symbols),
+                self._loop).result()
         for symbol in symbols:
             del self._handlers['lulds'][symbol]
 
@@ -406,7 +449,8 @@ class DataStream(_DataStream):
 
     def unregister_handler(self, msg_type, *symbols):
         for symbol in symbols:
-            del self._handlers[msg_type][symbol]
+            if symbol in self._handlers[msg_type]:
+                del self._handlers[msg_type][symbol]
 
 
 class CryptoDataStream(_DataStream):
@@ -415,7 +459,8 @@ class CryptoDataStream(_DataStream):
                  secret_key: str,
                  base_url: URL,
                  raw_data: bool,
-                 exchanges: Optional[List[str]] = None):
+                 exchanges: Optional[List[str]] = None,
+                 websocket_params: Optional[Dict] = None):
         self._key_id = key_id
         self._secret_key = secret_key
         base_url = re.sub(r'^http', 'ws', base_url)
@@ -429,16 +474,73 @@ class CryptoDataStream(_DataStream):
                          key_id=key_id,
                          secret_key=secret_key,
                          raw_data=raw_data,
+                         websocket_params=websocket_params,
                          )
+        self._handlers['orderbooks'] = {}
         self._name = 'crypto data'
 
+    def _cast(self, msg_type, msg):
+        result = super()._cast(msg_type, msg)
+        if not self._raw_data:
+            if msg_type == 'o':
+                result = OrderbookV2({
+                    orderbook_mapping_v2[k]: v
+                    for k, v in msg.items() if k in orderbook_mapping_v2
+                })
+        return result
+
+    async def _dispatch(self, msg):
+        msg_type = msg.get('T')
+        symbol = msg.get('S')
+        if msg_type == 'o':
+            handler = self._handlers['orderbooks'].get(
+                symbol, self._handlers['orderbooks'].get('*', None))
+            if handler:
+                await handler(self._cast(msg_type, msg))
+        else:
+            await super()._dispatch(msg)
+
+    async def _unsubscribe(self,
+                           trades=(),
+                           quotes=(),
+                           orderbooks=(),
+                           bars=(),
+                           updated_bars=(),
+                           daily_bars=()):
+        if (
+            trades or quotes or orderbooks or bars or updated_bars
+            or daily_bars
+        ):
+            await self._ws.send(
+                msgpack.packb({
+                    'action':      'unsubscribe',
+                    'trades':      trades,
+                    'quotes':      quotes,
+                    'orderbooks':  orderbooks,
+                    'bars':        bars,
+                    'updatedBars': updated_bars,
+                    'dailyBars':   daily_bars,
+                }))
+
+    def subscribe_orderbooks(self, handler, *symbols):
+        self._subscribe(handler, symbols, self._handlers['orderbooks'])
+
+    def unsubscribe_orderbooks(self, *symbols):
+        if self._running:
+            asyncio.run_coroutine_threadsafe(
+                self._unsubscribe(orderbooks=symbols),
+                self._loop).result()
+        for symbol in symbols:
+            del self._handlers['orderbooks'][symbol]
+
 
 class NewsDataStream(_DataStream):
     def __init__(self,
                  key_id: str,
                  secret_key: str,
                  base_url: URL,
-                 raw_data: bool):
+                 raw_data: bool,
+                 websocket_params: Optional[Dict] = None):
         self._key_id = key_id
         self._secret_key = secret_key
         base_url = re.sub(r'^http', 'ws', base_url)
@@ -447,6 +549,7 @@ class NewsDataStream(_DataStream):
                          key_id=key_id,
                          secret_key=secret_key,
                          raw_data=raw_data,
+                         websocket_params=websocket_params
                          )
         self._handlers = {
             'news':    {},
@@ -475,7 +578,7 @@ class NewsDataStream(_DataStream):
         if news:
             await self._ws.send(
                 msgpack.packb({
-                    'action':    'unsubscribe',
+                    'action': 'unsubscribe',
                     'news':    news,
                 }))
 
@@ -484,8 +587,9 @@ class NewsDataStream(_DataStream):
 
     def unsubscribe_news(self, *symbols):
         if self._running:
-            asyncio.get_event_loop().run_until_complete(
-                self._unsubscribe(news=symbols))
+            asyncio.run_coroutine_threadsafe(
+                self._unsubscribe(news=symbols),
+                self._loop).result()
         for symbol in symbols:
             del self._handlers['news'][symbol]
 
@@ -495,7 +599,8 @@ class TradingStream:
                  key_id: str,
                  secret_key: str,
                  base_url: URL,
-                 raw_data: bool = False):
+                 raw_data: bool = False,
+                 websocket_params: Optional[Dict] = None):
         self._key_id = key_id
         self._secret_key = secret_key
         base_url = re.sub(r'^http', 'ws', base_url)
@@ -503,12 +608,20 @@ class TradingStream:
         self._trade_updates_handler = None
         self._ws = None
         self._running = False
+        self._loop = None
         self._raw_data = raw_data
         self._stop_stream_queue = queue.Queue()
         self._should_run = True
+        self._websocket_params = websocket_params
+
+        if self._websocket_params is None:
+            self._websocket_params = WEBSOCKET_DEFAULTS
 
     async def _connect(self):
-        self._ws = await websockets.connect(self._endpoint)
+        self._ws = await websockets.connect(
+            self._endpoint,
+            **self._websocket_params
+        )
 
     async def _auth(self):
         await self._ws.send(
@@ -550,8 +663,9 @@ class TradingStream:
         _ensure_coroutine(handler)
         self._trade_updates_handler = handler
         if self._running:
-            asyncio.get_event_loop().run_until_complete(
-                self._subscribe_trade_updates())
+            asyncio.run_coroutine_threadsafe(
+                self._subscribe_trade_updates(),
+                self._loop).result()
 
     async def _start_ws(self):
         await self._connect()
@@ -577,6 +691,7 @@ class TradingStream:
                     pass
 
     async def _run_forever(self):
+        self._loop = asyncio.get_running_loop()
         # do not start the websocket connection until we subscribe to something
         while not self._trade_updates_handler:
             if not self._stop_stream_queue.empty():
@@ -590,7 +705,7 @@ class TradingStream:
             try:
                 if not self._should_run:
                     log.info("Trading stream stopped")
-                    break
+                    return
                 if not self._running:
                     log.info("starting trading websocket connection")
                     await self._start_ws()
@@ -618,6 +733,12 @@ class TradingStream:
         if self._stop_stream_queue.empty():
             self._stop_stream_queue.put_nowait({"should_stop": True})
 
+    def stop(self):
+        if self._loop.is_running():
+            asyncio.run_coroutine_threadsafe(
+                self.stop_ws(),
+                self._loop).result()
+
 
 class Stream:
     def __init__(self,
@@ -627,29 +748,34 @@ class Stream:
                  data_stream_url: URL = None,
                  data_feed: str = 'iex',
                  raw_data: bool = False,
-                 crypto_exchanges: Optional[List[str]] = None):
+                 crypto_exchanges: Optional[List[str]] = None,
+                 websocket_params: Optional[Dict] = None):
         self._key_id, self._secret_key, _ = get_credentials(key_id, secret_key)
         self._base_url = base_url or get_base_url()
-        self._data_steam_url = data_stream_url or get_data_stream_url()
+        self._data_stream_url = data_stream_url or get_data_stream_url()
 
         self._trading_ws = TradingStream(self._key_id,
                                          self._secret_key,
                                          self._base_url,
-                                         raw_data)
+                                         raw_data,
+                                         websocket_params=websocket_params)
         self._data_ws = DataStream(self._key_id,
                                    self._secret_key,
-                                   self._data_steam_url,
+                                   self._data_stream_url,
                                    raw_data,
-                                   data_feed.lower())
+                                   data_feed.lower(),
+                                   websocket_params=websocket_params)
         self._crypto_ws = CryptoDataStream(self._key_id,
                                            self._secret_key,
-                                           self._data_steam_url,
+                                           self._data_stream_url,
                                            raw_data,
-                                           crypto_exchanges)
+                                           crypto_exchanges,
+                                           websocket_params=websocket_params)
         self._news_ws = NewsDataStream(self._key_id,
                                        self._secret_key,
-                                       self._data_steam_url,
-                                       raw_data)
+                                       self._data_stream_url,
+                                       raw_data,
+                                       websocket_params=websocket_params)
 
     def subscribe_trade_updates(self, handler):
         self._trading_ws.subscribe_trade_updates(handler)
@@ -675,6 +801,9 @@ class Stream:
     def subscribe_bars(self, handler, *symbols):
         self._data_ws.subscribe_bars(handler, *symbols)
 
+    def subscribe_updated_bars(self, handler, *symbols):
+        self._data_ws.subscribe_updated_bars(handler, *symbols)
+
     def subscribe_daily_bars(self, handler, *symbols):
         self._data_ws.subscribe_daily_bars(handler, *symbols)
 
@@ -693,9 +822,15 @@ class Stream:
     def subscribe_crypto_bars(self, handler, *symbols):
         self._crypto_ws.subscribe_bars(handler, *symbols)
 
+    def subscribe_crypto_updated_bars(self, handler, *symbols):
+        self._crypto_ws.subscribe_updated_bars(handler, *symbols)
+
     def subscribe_crypto_daily_bars(self, handler, *symbols):
         self._crypto_ws.subscribe_daily_bars(handler, *symbols)
 
+    def subscribe_crypto_orderbooks(self, handler, *symbols):
+        self._crypto_ws.subscribe_orderbooks(handler, *symbols)
+
     def subscribe_news(self, handler, *symbols):
         self._news_ws.subscribe_news(handler, *symbols)
 
@@ -724,6 +859,13 @@ class Stream:
 
         return decorator
 
+    def on_updated_bar(self, *symbols):
+        def decorator(func):
+            self.subscribe_updated_bars(func, *symbols)
+            return func
+
+        return decorator
+
     def on_daily_bar(self, *symbols):
         def decorator(func):
             self.subscribe_daily_bars(func, *symbols)
@@ -752,7 +894,7 @@ class Stream:
 
         return decorator
 
-    def on_corrections(self, *symbols):
+    def on_correction(self, *symbols):
         def decorator(func):
             self._data_ws.register_handler("corrections", func, *symbols)
             return func
@@ -780,6 +922,13 @@ class Stream:
 
         return decorator
 
+    def on_crypto_updated_bar(self, *symbols):
+        def decorator(func):
+            self.subscribe_crypto_updated_bars(func, *symbols)
+            return func
+
+        return decorator
+
     def on_crypto_daily_bar(self, *symbols):
         def decorator(func):
             self.subscribe_crypto_daily_bars(func, *symbols)
@@ -787,6 +936,13 @@ class Stream:
 
         return decorator
 
+    def on_crypto_orderbook(self, *symbols):
+        def decorator(func):
+            self.subscribe_crypto_orderbooks(func, *symbols)
+            return func
+
+        return decorator
+
     def on_news(self, *symbols):
         def decorator(func):
             self.subscribe_news(func, *symbols)
@@ -805,6 +961,9 @@ class Stream:
     def unsubscribe_bars(self, *symbols):
         self._data_ws.unsubscribe_bars(*symbols)
 
+    def unsubscribe_updated_bars(self, *symbols):
+        self._data_ws.unsubscribe_updated_bars(*symbols)
+
     def unsubscribe_daily_bars(self, *symbols):
         self._data_ws.unsubscribe_daily_bars(*symbols)
 
@@ -823,9 +982,15 @@ class Stream:
     def unsubscribe_crypto_bars(self, *symbols):
         self._crypto_ws.unsubscribe_bars(*symbols)
 
+    def unsubscribe_crypto_updated_bars(self, *symbols):
+        self._crypto_ws.unsubscribe_updated_bars(*symbols)
+
     def unsubscribe_crypto_daily_bars(self, *symbols):
         self._crypto_ws.unsubscribe_daily_bars(*symbols)
 
+    def unsubscribe_crypto_orderbooks(self, *symbols):
+        self._crypto_ws.unsubscribe_orderbooks(*symbols)
+
     def unsubscribe_news(self, *symbols):
         self._news_ws.unsubscribe_news(*symbols)
 
@@ -836,9 +1001,8 @@ class Stream:
                              self._news_ws._run_forever())
 
     def run(self):
-        loop = asyncio.get_event_loop()
         try:
-            loop.run_until_complete(self._run_forever())
+            asyncio.run(self._run_forever())
         except KeyboardInterrupt:
             print('keyboard interrupt, bye')
             pass
@@ -859,13 +1023,23 @@ class Stream:
         if self._news_ws:
             await self._news_ws.stop_ws()
 
+    def stop(self):
+        if self._trading_ws:
+            self._trading_ws.stop()
+        if self._data_ws:
+            self._data_ws.stop()
+        if self._crypto_ws:
+            self._crypto_ws.stop()
+        if self._news_ws:
+            self._news_ws.stop()
+
     def is_open(self):
         """
         Checks if either of the websockets is open
         :return:
         """
         open_ws = (self._trading_ws._ws or self._data_ws._ws
-                   or self._crypto_ws._ws or self._news_ws) # noqa
+                   or self._crypto_ws._ws or self._news_ws)  # noqa
         if open_ws:
             return True
         return False
